
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/random.h>
#include <errno.h>

#define MAX_USERS  0x10000

#define UUIDLEN 0x10
#define UUID_BUFFER_LEN (UUIDLEN*2+5)

#define KEY_LENGTH 0x20

/* Maximum supported 25 MB entries */
#define MAX_ENTRY_SIZE (25*1024*1024)

#define ERROR(msg) do { fputs(msg, stderr); exit(1); } while (0)


/* Generic doubly-linked list implementation
 * - taken after the linux kernel list implementation */
typedef struct __list {
	struct __list* next;
	struct __list* prev;
} list;

typedef struct {
	list user_list;
	unsigned char* uuid;
	char* name;
	list entries;
	size_t num_entries;
} user_t;

typedef struct __entry {
	list entry_list;
	size_t length;
	unsigned char data[];
} entry_t;

typedef struct {
	list current_users;
	list* free_users;
	user_t users[MAX_USERS];
	unsigned char master_key[KEY_LENGTH];
} state_t;

state_t state;




void encrypt_data(unsigned char* key, const unsigned char* plaintext, unsigned char* ciphertext, size_t plaintext_length) {
	for(size_t i = 0; i < plaintext_length; ++i)
		ciphertext[i] = plaintext[i] ^ key[i % KEY_LENGTH];
}

bool decrypt_data(unsigned char* key, const unsigned char* ciphertext, unsigned char* plaintext, size_t ciphertext_length) {
	for(size_t i = 0; i < ciphertext_length; ++i)
		plaintext[i] = ciphertext[i] ^ key[i % KEY_LENGTH];
	return true;
}



/*
 *************************************************************************************************
 *************************************** BEGIN SAFE BLOCK ****************************************
 *           To save you some time in the qualification, the functions below are marked
 *      as safe, i.e. without bugs and vulnerabilities. Note, this does not mean that all other
 *                         functions contain bugs or vulnerabilities
 *************************************************************************************************
 *************************************************************************************************
 */

#define FOR_ALL_USERS(state, it) \
	for(list* _i = (state)->current_users.next; \
	     _i != &(state)->current_users && \
          ((it) = list_to_user(_i), 1); \
	     _i = _i->next)

/* This macro returns the "container" of a substruct and is used to return the student for a specific list */
#define containerof(ptr, type, field) ((type*)((void*)ptr - offsetof(type, field)))


/** These functions can be used within gdb to inspect the data structures **/
/* Example1: Dumping information for the first user in current_users:
 *  print *list_to_user(state.current_users.next)
 * Example2: Dumping information for the second user in current_users:
 *  print *list_to_user(state.current_users.next.next)
 * Note, GDB uses C-Syntax here and the asterisk dereferences the pointer (and therefore GDB shows all fields) */
user_t* list_to_user(list* l) {
	return containerof(l, user_t, user_list);
}

/* Example: Dumping information for the first entry of the first user in current_users:
 *  print *list_to_entry(list_to_user(state.current_users.next)->entries.next)
 * Note, GDB uses C-Syntax here and the asterisk dereferences the pointer (and therefore GDB shows all fields) */
entry_t* list_to_entry(list* l) {
	return containerof(l, entry_t, entry_list);
}

const char* uuid_format = "%02hhx%02hhx%02hhx%02hhx-%02hhx%02hhx-%02hhx%02hhx-%02hhx%02hhx-%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx";

void bytes2uuid(const unsigned char* bytes, char* txt) {
	sprintf(txt, uuid_format,
			bytes[ 0], bytes[ 1], bytes[ 2], bytes[ 3],
			bytes[ 4], bytes[ 5], bytes[ 6], bytes[ 7],
			bytes[ 8], bytes[ 9], bytes[10], bytes[11],
			bytes[12], bytes[13], bytes[14], bytes[15]);
}

bool uuid2bytes(const char* txt, unsigned char* bytes) {
	int r = sscanf(txt, uuid_format,
			&bytes[ 0], &bytes[ 1], &bytes[ 2], &bytes[ 3],
			&bytes[ 4], &bytes[ 5], &bytes[ 6], &bytes[ 7],
			&bytes[ 8], &bytes[ 9], &bytes[10], &bytes[11],
			&bytes[12], &bytes[13], &bytes[14], &bytes[15]);
	return r == 16;
}



void init_challenge(void) {
	int fd;
	unsigned int rand_val;
	// setting stdout and stdin to be unbuffered
	setvbuf(stdout,0,_IONBF,0);
	setvbuf(stdin,0,_IONBF,0);
	// seeding the random number generator
	fd = open("/dev/urandom", O_RDONLY);
	read(fd, &rand_val, sizeof(rand_val));
	close(fd);
	srand(rand_val);
}


void list_init_head(list* list) {
	list->next = list;
	list->prev = list;
}

void list_remove(list* entry) {
	entry->prev->next = entry->next;
	entry->next->prev = entry->prev;
}

void list_add(list* head, list* entry) {
	entry->next = head->next;
	entry->prev = head;

	head->next->prev = entry;
	head->next = entry;
}

void list_replace(list* eold, list* enew) {
	enew->next = eold->next;
	enew->prev = eold->prev;

	enew->next->prev = enew;
	enew->prev->next = enew;

	eold->next = eold;
	eold->prev = eold;
}


user_t* alloc_user(state_t *state) {
	if(!state->free_users) return NULL;
	list* rl = state->free_users;
	state->free_users = rl->next;
	user_t* u = list_to_user(rl);
	u->uuid = malloc(UUIDLEN);
	return u;
}


void add_to_freelist(list** head, list* entry) {
	entry->next = *head;
	*head = entry;
}

user_t* get_user_by_uuid(state_t* state, unsigned char* uuid) {
	user_t* u;
	FOR_ALL_USERS(state, u) {
		if(!memcmp(u->uuid, uuid, UUIDLEN))
			return u;
	}
	return NULL;
}


entry_t* prepare_entry(state_t* state, unsigned char* data, size_t datalength) {
	entry_t* n = calloc(1, datalength+sizeof(entry_t));
	n->length = datalength;
	encrypt_data(state->master_key, data, n->data, datalength);
	return n;
}




void init_state(state_t* state) {
	getrandom(state->master_key, sizeof(state->master_key), 0);

	state->free_users = NULL;
	for(ssize_t i = 0; i < MAX_USERS; ++i)
		add_to_freelist(&state->free_users, &state->users[i].user_list);

	list_init_head(&state->current_users);
}

bool read_exactly(int fd, void* buf, size_t count) {
	while(count) {
		ssize_t r = read(fd, buf, count);
		if (r < 0) {
			if(errno == EINTR)
				continue;
			return false;
		}
		if (r == 0)
			return false;

		buf += r;
		count -= r;
	}
	return true;
}


char get_input(void) {
	char choice[0x20];
	printf(" > ");
	if(!fgets(choice, sizeof(choice), stdin)) ERROR("Could not read input...");
	return choice[0];
}

ssize_t get_int(void) {
	char choice[0x20];
	printf(" > ");
	if(!fgets(choice, sizeof(choice), stdin)) ERROR("Could not read input...");
	return atoll(choice);
}

char* _getline(void) {
	static char* buf = NULL;
	static size_t n = 0;
	if(getline(&buf, &n, stdin) < 0) ERROR("Could not read line...");
	return buf;
}

bool get_data(unsigned char** rbuf, size_t* rlen) {
	printf("Data length?");
	ssize_t dlen = get_int();
	if(dlen <= 0 || dlen > MAX_ENTRY_SIZE) {
		puts("Invalid length!");
		return false;
	}

	unsigned char* buf = malloc(dlen);
	puts("Data:");
	if(!read_exactly(0, buf, dlen)) {
		puts("Error reading...");
		exit(1);
	}
	*rbuf = buf;
	*rlen = dlen;
	return true;
}




/*
 *************************************************************************************************
 **************************************** END SAFE BLOCK *****************************************
 *           To save you some time in the qualification, the functions above are marked
 *      as safe, i.e. without bugs and vulnerabilities. Note, this does not mean that all other
 *                         functions contain bugs or vulnerabilities.
 *************************************************************************************************
 *************************************************************************************************
 */


void print_hello(void) {
	puts("Welcome to the pilot phase of the secure digital user record platform!");
	puts("After fixing the 21 vulnerabilities found by a previous security audit, we can proclaim us to be finally secure!");
	puts("We're using state-of-the-art encryption to make sure, only YOU (and all other necessary parties) can access the data!");
	puts("After registration you will be assigned a UUID that can be used to access your data!");
}


int main_menu(void) {
	puts("=================================");
	puts(" What would you like to do ?");
	puts(" (L)ogin");
	puts(" (R)egister");
	puts(" e(x)it");
	puts("=================================");
	return get_input();
}

int user_menu(void) {
	puts("=================================");
	puts(" What would you like to do ?");
	puts(" (A)dd entry");
	puts(" (R)ead entry");
	puts(" (M)odify entry");
	puts(" (D)elete entry");
	puts(" (L)ogout");
	puts("=================================");
	return get_input();
}



entry_t* new_entry(state_t* state, user_t* user,
					unsigned char* data, size_t datalength) {
	entry_t* n = prepare_entry(state, data, datalength);
	list_add(&user->entries, &n->entry_list);
	user->num_entries += 1;
	return n;
}


void delete_entry(entry_t* entry) {
	list_remove(&entry->entry_list);
	free(entry);
}


user_t* new_user(state_t *state, const char* name) {
	user_t* user = alloc_user(state);
	if(!user) return NULL;

	getrandom(user->uuid, UUIDLEN, 0);
	user->name = strdup(name);
	list_init_head(&user->entries);
	list_add(&state->current_users, &user->user_list);
	return user;
}

void print_entry(entry_t* e, bool decrypt) {
	if(decrypt) {
		unsigned char* buf = calloc(1, e->length);
		if(!buf) {
			puts("Out of memory, cannot print such a large entry!");
			return;
		}
		if(!decrypt_data(state.master_key, e->data, buf, e->length)) {
			puts("Decryption failed, data CORRUPTED!");
			return;
		}
		puts("===BEGIN ENTRY DATA===");
		fflush(stdout);
		write(1, buf, e->length);
		puts("===END ENTRY DATA===");
		free(buf);
	} else {
		puts("===BEGIN ENCRYPTED ENTRY DATA===");
		fflush(stdout);
		write(1, e->data, e->length);
		puts("===END ENTRY DATA===");
	}
}



entry_t* select_entry(user_t* user, const char* what)
{
		size_t count = user->num_entries;
		if(!count) {
			puts("There are no entries for you here...");
			return NULL;
		}
		printf("There are %zu entries, which one would you like to %s? [1..%zu]\n", count, what, count);
		ssize_t entry_id = get_int();
		if(entry_id < 1 || (size_t)entry_id > count) {
			puts("Invalid index");
			return NULL;
		}

		list* i;
		for(i = user->entries.next; entry_id > 1; i = i->next)
			--entry_id;

		return list_to_entry(i);
}


void do_read_entry(user_t* user)
{
	entry_t* entry = select_entry(user, "read");
	if(!entry) return;
	printf("Decrypt [Y/N]?");
	char choice = get_input();
	bool decrypt = !(choice == 'N' || choice == 'n');
	print_entry(entry, decrypt);
}

void do_modify_entry(user_t* user)
{
	entry_t* entry = select_entry(user, "modify");
	if(!entry) return;
	unsigned char* buf; size_t len;
	if(!get_data(&buf, &len)) return;

	/* If same size, we can replace in-place */
	if (entry->length == len) {
		encrypt_data(state.master_key, buf, entry->data, len);
	}
	/* Else we allocate and replace */
	else {
		entry_t* enew = prepare_entry(&state, buf, len);
		list_replace(&entry->entry_list, &enew->entry_list);
		free(entry);
	}
}





void do_user(user_t* user) {
	while(1) {
		switch(user_menu()) {
			case 'a':
			case 'A':
			{
				unsigned char* buf; size_t dlen;

				if(!get_data(&buf, &dlen)) break;
				new_entry(&state, user, buf, dlen);
				break;
			}

			case 'r':
			case 'R':
				do_read_entry(user);
				break;

			case 'm':
			case 'M':
				do_modify_entry(user);
				break;

			case 'D':
			case 'd':
			{
				entry_t* entry = select_entry(user, "delete");
				if(!entry) break;

				delete_entry(entry);
				puts("Entry deleted!");
				break;
			}

			case 'l':
			case 'L':
				puts("Logged out successfully...");
				return;
		}
	}
}




void login(void) {
	printf("UUID: ");
	char* line = _getline();
	unsigned char uuid[UUID_BUFFER_LEN];

	if(!uuid2bytes(line, uuid)) {
		puts("Invalid UUID");
		return;
	}

	user_t* u;
	FOR_ALL_USERS(&state, u) {
		if(!memcmp(u->uuid, uuid, UUIDLEN))
		{
			do_user(u);
			return;
		}
	}

	puts("UUID not found!");
	return;
}

void register_user(void) {
	char uuidbuf[UUID_BUFFER_LEN];
	printf("What's your name? > ");
	char* name = _getline();
	user_t* u = new_user(&state, name);
	if(u) {
		bytes2uuid(u->uuid, uuidbuf);
		printf("Your UUID is: %s\n", uuidbuf);
	}
	else
		puts("Out of space, ask the administrator to clean up!");
}


void init_mock_database(void) {
	char flag[0x100] = { 0 };

	FILE* fp;
	if(!(fp = popen("get_flag", "r")))
		ERROR("Could not read flag");
	fgets(flag, sizeof(flag), fp);
	pclose(fp);
	size_t l = strlen(flag);

	user_t* target = new_user(&state, "target");
	new_entry(&state, target, (unsigned char*)flag, l);
}


int main() {
	init_challenge();
	print_hello();
	init_state(&state);
	init_mock_database();

	while(1) {
		switch(main_menu()) {
			case 'l':
			case 'L':
				login();
				break;

			case 'r':
			case 'R':
				register_user();
				break;

			case 'x':
			case 'X':
				return 0;

			default:
				puts("Invalid input!");
				break;
		}
	}
}

